from __future__ import annotations

from pathlib import Path
from typing import ClassVar

import orjson as json
from cryptography.fernet import Fernet, InvalidToken


class PantherDBException(Exception):
    pass


class PantherDB:
    _instances: ClassVar[dict] = {}
    db_name: str = 'database.pdb'
    __secret_key: bytes | None
    __fernet: Fernet | None
    __return_dict: bool
    __content: dict

    def __new__(cls, *args, **kwargs):
        if cls.__name__ != 'PantherDB':
            return super().__new__(cls)

        if args:
            db_name = args[0]
        elif 'db_name' in kwargs:
            db_name = kwargs['db_name']
        else:
            db_name = cls.db_name

        db_name = db_name.removesuffix('.pdb')
        if db_name not in cls._instances:
            cls._instances[db_name] = super().__new__(cls)
        return cls._instances[db_name]

    def __init__(
            self,
            db_name: str | None = None,
            *,
            return_dict: bool = False,
            secret_key: bytes | None = None,
    ):
        self.__return_dict = return_dict
        self.__secret_key = secret_key
        self.__fernet = Fernet(self.__secret_key) if self.__secret_key else None
        self.__content = {}

        if db_name:
            if not db_name.endswith('pdb'):
                db_name = f'{db_name}.pdb'
            self.db_name = db_name
        Path(self.db_name).touch(exist_ok=True)

    def __str__(self) -> str:
        self._refresh()
        db = ',\n'.join(f'    {k}: {len(v)} documents' for k, v in self.content.items())
        return f'{self.__class__.__name__}(\n{db}\n)'

    __repr__ = __str__

    @property
    def content(self) -> dict:
        return self.__content

    @property
    def return_dict(self) -> bool:
        return self.__return_dict

    @property
    def secret_key(self) -> bytes | None:
        return self.__secret_key

    def _write(self) -> None:
        content = json.dumps(self.content)

        if self.secret_key:
            content = self.__fernet.encrypt(content)

        with open(self.db_name, 'bw') as file:
            file.write(content)

    def _refresh(self) -> None:
        # TODO: Find Solution, so won't refresh on every single query
        with open(self.db_name, 'rb') as file:

            if not (data := file.read()):
                self.__content = {}

            elif not self.secret_key:
                self.__content = json.loads(data)

            else:
                try:
                    decrypted_data: bytes = self.__fernet.decrypt(data)
                    self.__content = json.loads(decrypted_data)
                except InvalidToken:
                    error = '"secret_key" Is Not Valid'
                    raise PantherDBException(error)

    def collection(self, collection_name: str) -> PantherCollection:
        return PantherCollection(
            db_name=self.db_name,
            collection_name=collection_name,
            return_dict=self.return_dict,
            secret_key=self.secret_key,
        )

    def close(self):
        pass


class PantherCollection(PantherDB):
    __collection_name: str

    def __init__(
            self,
            db_name: str,
            *,
            collection_name: str,
            return_dict: bool,
            secret_key: bytes,
    ):
        super().__init__(db_name=db_name, return_dict=return_dict, secret_key=secret_key)
        self.__collection_name = collection_name

    def __str__(self) -> str:
        self._refresh()

        if self.collection_name not in self.content or (documents := self.content[self.collection_name]):
            result = ''
        else:
            result = '\n'.join(f'    {k}: {type(v).__name__}' for k, v in documents[0].items())

        return f'{self.__class__.__name__}(\n    collection_name: {self.collection_name}\n\n{result}\n)'

    @property
    def collection_name(self) -> str:
        return self.__collection_name

    def __check_is_panther_document(self) -> None:
        if not hasattr(self, '_id'):
            raise PantherDBException('You should call this method on PantherDocument instance.')

    def __create_result(self, data: dict, /) -> PantherDocument | dict:
        if self.return_dict:
            return data
        else:
            return PantherDocument(
                db_name=self.db_name,
                collection_name=self.collection_name,
                return_dict=self.return_dict,
                secret_key=self.secret_key,
                **data,
            )

    def _write_collection(self, documents: list) -> None:
        self.content[self.collection_name] = documents
        self._write()

    def _drop_collection(self) -> None:
        self._refresh()
        if self.collection_name in self.content:
            del self.content[self.collection_name]
        self._write()

    def _get_collection(self) -> list[dict]:
        """Return documents"""
        self._refresh()
        return self.content.get(self.collection_name, [])

    def find_one(self, **kwargs) -> PantherDocument | dict | None:
        documents = self._get_collection()

        # Empty Collection
        if not documents:
            return None

        if not kwargs:
            if self.return_dict:
                return documents[0]
            else:
                return self.__create_result(documents[0])

        for d in documents:
            for k, v in kwargs.items():
                if d.get(k) != v:
                    break
            else:
                return self.__create_result(d)

    def find(self, **kwargs) -> list[PantherDocument | dict]:
        if not kwargs:
            return self.all()

        result = list()
        for r in self._get_collection():
            for k, v in kwargs.items():
                if r.get(k) != v:
                    break
            else:
                result.append(self.__create_result(r))
        return result

    def first(self, **kwargs) -> PantherDocument | dict | None:
        return self.find_one(**kwargs)

    def last(self, **kwargs) -> PantherDocument | dict | None:
        if result := self.find(**kwargs):
            return result[-1]
        return None

    def all(self) -> list[PantherDocument | dict]:
        if self.return_dict:
            return self._get_collection()
        else:
            return [self.__create_result(r) for r in self._get_collection()]

    def insert_one(self, **kwargs) -> PantherDocument | dict:
        documents = self._get_collection()
        kwargs['_id'] = len(documents) + 1
        documents.append(kwargs)
        self._write_collection(documents)
        return self.__create_result(kwargs)

    def delete(self) -> None:
        self.__check_is_panther_document()
        documents = self._get_collection()
        for d in documents:
            if d.get('_id') == self._id:  # noqa: Unresolved References
                documents.remove(d)
                self._write_collection(documents)
                break

    def delete_one(self, **kwargs) -> bool:
        documents = self._get_collection()
        if not kwargs:
            return False

        index = 0
        found = False
        for d in documents:
            for k, v in kwargs.items():
                if d.get(k) != v:
                    break
            else:
                found = True
                break
            index += 1

        # Didn't find any match
        if not found:
            return False

        # Delete Matched One
        documents.pop(index)
        self._write_collection(documents)
        return True

    def delete_many(self, **kwargs) -> int:
        documents = self._get_collection()
        if not kwargs:
            return 0

        indexes = list()
        index = 0
        found = False
        for d in documents:
            for k, v in kwargs.items():
                if d.get(k) != v:
                    break
            else:
                indexes.append(index)
                found = True
            index += 1

        # Didn't find any match
        if not found:
            return 0

        # Delete Matched Indexes
        for i in indexes[::-1]:
            documents.pop(i)
        self._write_collection(documents)
        return len(indexes)

    def update(self, **kwargs) -> None:
        self.__check_is_panther_document()
        documents = self._get_collection()
        for d in documents:
            if d.get('_id') == self._id:  # noqa: Unresolved References
                for k, v in kwargs.items():
                    d[k] = v
                    setattr(self, k, v)
                self._write_collection(documents)

    def update_one(self, condition: dict, **kwargs) -> bool:
        documents = self._get_collection()
        result = False

        if not condition:
            return result

        for d in documents:
            for k, v in condition.items():
                if d.get(k) != v:
                    break
            else:
                result = True
                for new_k, new_v in kwargs.items():
                    d[new_k] = new_v
                self._write_collection(documents)
                break

        return result

    def update_many(self, condition: dict, **kwargs) -> int:
        documents = self._get_collection()
        if not condition:
            return 0

        updated_count = 0
        for d in documents:
            for k, v in condition.items():
                if d.get(k) != v:
                    break
            else:
                updated_count += 1
                for new_k, new_v in kwargs.items():
                    d[new_k] = new_v

        if updated_count:
            self._write_collection(documents)
        return updated_count

    def count(self, **kwargs) -> int:
        documents = self._get_collection()
        if not kwargs:
            return len(documents)

        result = 0
        for d in documents:
            for k, v in kwargs.items():
                if d.get(k) != v:
                    break
            else:
                result += 1
        return result

    def drop(self) -> None:
        self._drop_collection()


class PantherDocument(PantherCollection):
    __data: dict

    def __init__(
            self,
            db_name: str,
            *,
            collection_name: str,
            return_dict: bool,
            secret_key: bytes,
            **kwargs,
    ):
        self.__data = kwargs
        super().__init__(
            db_name=db_name,
            collection_name=collection_name,
            return_dict=return_dict,
            secret_key=secret_key,
        )

    def __str__(self) -> str:
        items = ', '.join(f'id={v}' if k == '_id' else f'{k}={v}' for k, v in self.data.items())
        return f'{self.collection_name}({items})'

    __repr__ = __str__

    def __getattr__(self, item: str):
        try:
            return object.__getattribute__(self, '_PantherDocument__data')[item]
        except KeyError:
            error = f'Invalid Collection Field: "{item}"'
            raise PantherDBException(error)

    def __setattr__(self, key, value):
        if key not in [
            '_PantherDB__return_dict',
            '_PantherDB__secret_key',
            '_PantherDB__content',
            '_PantherDB__fernet',
            '_PantherCollection__collection_name',
            '_PantherDocument__data',
        ]:
            try:
                object.__getattribute__(self, key)
            except AttributeError:
                self.data[key] = value
                return

        super().__setattr__(key, value)

    @property
    def id(self) -> int:
        return self.data['_id']

    @property
    def data(self) -> dict:
        return self.__data

    def save(self) -> None:
        """Pop & Insert New Document"""
        documents = self._get_collection()
        for i, d in enumerate(documents):
            if d['_id'] == self.id:
                documents.pop(i)
                documents.insert(i, self.data)
                break
        self._write_collection(documents)

    def json(self) -> str:
        return json.dumps(self.data).decode()




from pathlib import Path
from unittest import TestCase
from uuid import uuid4

import orjson as json
from faker import Faker

from pantherdb import PantherCollection, PantherDB, PantherDocument

f = Faker()


class TestNormalPantherDB(TestCase):

    @classmethod
    def setUp(cls):
        cls.db_name = uuid4().hex
        cls.db_name = f'{cls.db_name}.pdb'
        cls.db = PantherDB(db_name=cls.db_name)

    @classmethod
    def tearDown(cls):
        Path(cls.db_name).unlink()

    @classmethod
    def create_junk_document(cls, collection) -> int:
        _count = f.random.randint(2, 10)
        for i in range(_count):
            collection.insert_one(first_name=f'{f.first_name()}{i}', last_name=f'{f.last_name()}{i}')
        return _count

    # Singleton
    def test_pantherdb_singleton(self):
        test_1 = PantherDB(db_name='test1')
        test_2 = PantherDB('test1')
        self.assertEqual(test_1, test_2)

        default_1 = PantherDB()
        default_2 = PantherDB()
        self.assertEqual(default_1, default_2)

        self.assertNotEqual(test_1, default_1)
        self.assertNotEqual(test_2, default_2)

        Path(test_1.db_name).unlink()
        Path(default_1.db_name).unlink()

    # Create DB
    def test_creation_of_db(self):
        self.assertTrue(Path(self.db_name).exists())
        self.assertTrue(Path(self.db_name).is_file())
        self.assertEqual(self.db.db_name, self.db_name)

    def test_creation_of_db_without_extension(self):
        db_name = uuid4().hex
        db = PantherDB(db_name=db_name)
        final_db_name = f'{db_name}.pdb'

        self.assertTrue(Path(final_db_name).exists())
        self.assertTrue(Path(final_db_name).is_file())
        self.assertEqual(db.db_name, final_db_name)

        Path(final_db_name).unlink()

    def test_creation_of_collection(self):
        collection_name = f.word()
        collection = self.db.collection(collection_name)

        self.assertTrue(bool(collection))
        self.assertTrue(isinstance(collection, PantherCollection))
        self.assertEqual(collection.collection_name, collection_name)
        self.assertEqual(collection.content, {})
        self.assertEqual(collection.secret_key, None)

    # Drop
    def test_drop_collection(self):
        collection = self.db.collection(f.word())
        collection.drop()
        self.assertEqual(collection.content, {})

    # Insert
    def test_insert_one(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()
        obj = collection.insert_one(first_name=first_name, last_name=last_name)

        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.id, 1)
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)

    def test_id_assignments(self):
        collection = self.db.collection(f.word())
        _count = f.random.randint(2, 10)
        for i in range(_count):
            obj = collection.insert_one(first_name=f.first_name(), last_name=f.last_name())

        # obj will overwrite after every insert in loop, and we check the last obj
        self.assertEqual(obj.id, _count)

    # Find One
    def test_find_one_first(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.find_one(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)

    def test_find_one_last(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Add others
        self.create_junk_document(collection)

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Find
        obj = collection.find_one(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)

    def test_find_one_none(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.find_one(first_name=first_name, last_name=last_name)
        self.assertIsNone(obj)

    def test_find_one_with_kwargs_from_empty_collection(self):
        collection = self.db.collection(f.word())

        # Find
        obj = collection.find_one(first_name=f.first_name(), last_name=f.last_name())
        self.assertIsNone(obj)

    def test_find_one_without_kwargs_from_empty_collection(self):
        collection = self.db.collection(f.word())

        # Find
        obj = collection.find_one()
        self.assertIsNone(obj)

    # First
    def test_first_when_its_first(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.first(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)

    def test_first_of_many_finds(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        expected = collection.insert_one(first_name=first_name, last_name=last_name)
        collection.insert_one(first_name=first_name, last_name=last_name)
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.first(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.id, expected.id)

    def test_first_when_its_last(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Add others
        self.create_junk_document(collection)

        # Insert with specific names
        expected = collection.insert_one(first_name=first_name, last_name=last_name)

        # Find
        obj = collection.first(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)
        self.assertEqual(obj.id, expected.id)

    def test_first_none(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.first(first_name=first_name, last_name=last_name)
        self.assertIsNone(obj)

    def test_first_with_kwargs_from_empty_collection(self):
        collection = self.db.collection(f.word())

        # Find
        obj = collection.first(first_name=f.first_name(), last_name=f.last_name())
        self.assertIsNone(obj)

    def test_first_without_kwargs_from_empty_collection(self):
        collection = self.db.collection(f.word())

        # Find
        obj = collection.first()
        self.assertIsNone(obj)

    # Last
    def test_last_when_its_first(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.first(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)

    def test_last_of_many_finds(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)
        collection.insert_one(first_name=first_name, last_name=last_name)
        expected = collection.insert_one(first_name=first_name, last_name=last_name)

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.last(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.id, expected.id)

    def test_last_when_its_last(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Add others
        self.create_junk_document(collection)

        # Insert with specific names
        expected = collection.insert_one(first_name=first_name, last_name=last_name)

        # Find
        obj = collection.last(first_name=first_name, last_name=last_name)
        self.assertTrue(isinstance(obj, PantherDocument))
        self.assertEqual(obj.first_name, first_name)
        self.assertEqual(obj.last_name, last_name)
        self.assertEqual(obj.id, expected.id)

    def test_last_none(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Add others
        self.create_junk_document(collection)

        # Find
        obj = collection.last(first_name=first_name, last_name=last_name)
        self.assertIsNone(obj)

    def test_last_with_kwargs_from_empty_collection(self):
        collection = self.db.collection(f.word())

        # Find
        obj = collection.last(first_name=f.first_name(), last_name=f.last_name())
        self.assertIsNone(obj)

    def test_last_without_kwargs_from_empty_collection(self):
        collection = self.db.collection(f.word())

        # Find
        obj = collection.last()
        self.assertIsNone(obj)

    # Find
    def test_find_response_type(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find
        objs = collection.find(first_name=first_name)

        self.assertTrue(isinstance(objs, list))
        self.assertEqual(len(objs), 1)
        self.assertTrue(isinstance(objs[0], PantherDocument))

    def test_find_with_filter(self):
        collection = self.db.collection(f.word())

        # Add others
        self.create_junk_document(collection)

        # Insert with specific names
        first_name = f.first_name()
        _count = f.random.randint(2, 10)
        for i in range(_count):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find
        objs = collection.find(first_name=first_name)

        self.assertTrue(isinstance(objs, list))
        self.assertEqual(len(objs), _count)
        for i in range(_count):
            self.assertEqual(objs[i].first_name, first_name)

    def test_find_without_filter(self):
        collection = self.db.collection(f.word())

        # Add others
        _count_1 = self.create_junk_document(collection)

        # Insert with specific names
        first_name = f.first_name()
        _count_2 = f.random.randint(2, 10)
        for i in range(_count_2):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find
        objs = collection.find()
        _count_all = _count_1 + _count_2

        self.assertTrue(isinstance(objs, list))
        self.assertEqual(len(objs), _count_all)
        for i in range(_count_all):
            self.assertTrue(isinstance(objs[i], PantherDocument))

        # Check count of specific name
        specific_count = 0
        for i in range(_count_all):
            if objs[i].first_name == first_name:
                specific_count += 1

        self.assertEqual(specific_count, _count_2)

    def test_find_all(self):
        collection = self.db.collection(f.word())

        # Add others
        _count_1 = self.create_junk_document(collection)

        # Insert with specific names
        first_name = f.first_name()
        _count_2 = f.random.randint(2, 10)
        for i in range(_count_2):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find
        objs = collection.all()
        _count_all = _count_1 + _count_2

        self.assertTrue(isinstance(objs, list))
        self.assertEqual(len(objs), _count_all)
        for i in range(_count_all):
            self.assertTrue(isinstance(objs[i], PantherDocument))

        # Check count of specific name
        specific_count = 0
        for i in range(_count_all):
            if objs[i].first_name == first_name:
                specific_count += 1

        self.assertEqual(specific_count, _count_2)

    # Count
    def test_count_all(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        # Count them
        count_all = collection.count()

        self.assertEqual(count_all, _count)
        self.assertEqual(count_all, len(collection.all()))

    def test_count_with_filter(self):
        collection = self.db.collection(f.word())

        # Add others
        _count_1 = self.create_junk_document(collection)

        # Insert with specific names
        first_name = f.first_name()
        _count_2 = f.random.randint(2, 10)
        for i in range(_count_2):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        count_specific = collection.count(first_name=first_name)
        self.assertEqual(count_specific, _count_2)
        self.assertEqual(count_specific, len(collection.find(first_name=first_name)))

    # Delete Self
    def test_delete(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        # Insert with specific name
        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find
        obj = collection.find_one(first_name=first_name)

        # Delete It
        obj.delete()

        # Find It Again
        new_obj = collection.find_one(first_name=first_name)
        self.assertEqual(new_obj, None)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count)

    # Delete One
    def test_delete_one(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Delete One
        is_deleted = collection.delete_one(first_name=first_name)
        self.assertTrue(is_deleted)

        # Find It Again
        new_obj = collection.find_one(first_name=first_name)
        self.assertEqual(new_obj, None)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count)

    def test_delete_one_not_found(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        first_name = f.first_name()

        # Delete One
        is_deleted = collection.delete_one(first_name=first_name)
        self.assertFalse(is_deleted)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count)

    def test_delete_one_first(self):
        collection = self.db.collection(f.word())

        # Add others
        _count_1 = self.create_junk_document(collection)

        first_name = f.first_name()
        _count_2 = f.random.randint(2, 10)
        for i in range(_count_2):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Delete One
        is_deleted = collection.delete_one(first_name=first_name)
        self.assertTrue(is_deleted)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count_1 + _count_2 - 1)

        # Count of undeleted
        undeleted_count = collection.count(first_name=first_name)
        self.assertEqual(undeleted_count, _count_2 - 1)

    # Delete Many
    def test_delete_many(self):
        collection = self.db.collection(f.word())

        # Add others
        _count_1 = self.create_junk_document(collection)

        # Insert with specific name
        first_name = f.first_name()
        _count_2 = f.random.randint(2, 10)
        for i in range(_count_2):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Delete Many
        deleted_count = collection.delete_many(first_name=first_name)
        self.assertEqual(deleted_count, _count_2)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count_1)

    def test_delete_many_not_found(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        first_name = f.first_name()

        # Delete Many
        deleted_count = collection.delete_many(first_name=first_name)
        self.assertEqual(deleted_count, 0)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count)

    # Update Self
    def test_update(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        # Insert with specific name
        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find One
        obj = collection.find_one(first_name=first_name)
        new_name = f.first_name()
        obj.update(first_name=new_name)
        self.assertEqual(obj.first_name, new_name)

        # Find with old name
        old_obj = collection.find_one(first_name=first_name)
        self.assertIsNone(old_obj)

        # Find with new name
        obj = collection.find_one(first_name=new_name)
        self.assertEqual(obj.first_name, new_name)

    # Update One
    def test_update_one_single_document(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        # Insert with specific name
        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Update One
        new_name = f.first_name()
        is_updated = collection.update_one({'first_name': first_name}, first_name=new_name)
        self.assertTrue(is_updated)

        # Find with old name
        old_obj = collection.find_one(first_name=first_name)
        self.assertIsNone(old_obj)

        # Find with new name
        obj = collection.find_one(first_name=new_name)
        self.assertEqual(obj.first_name, new_name)

    def test_update_one_single_document_not_found(self):
        collection = self.db.collection(f.word())

        # Add others
        _count = self.create_junk_document(collection)

        # Insert with specific name
        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Update One
        new_name = f.first_name()
        is_updated = collection.update_one({'first_name': f.first_name()}, first_name=new_name)
        self.assertFalse(is_updated)

        # Find with old name
        old_obj = collection.find_one(first_name=first_name)
        self.assertIsNotNone(old_obj)

        # Find with new name
        obj = collection.find_one(first_name=new_name)
        self.assertIsNone(obj)

    # Update Many
    def test_update_many(self):
        collection = self.db.collection(f.word())

        # Add others
        _count_1 = self.create_junk_document(collection)

        # Insert with specific name
        first_name = f.first_name()
        _count_2 = f.random.randint(2, 10)
        for i in range(_count_2):
            collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Update Many
        new_name = f.first_name()
        updated_count = collection.update_many({'first_name': first_name}, first_name=new_name)
        self.assertEqual(updated_count, _count_2)

        # Find Them with old name
        objs = collection.find(first_name=first_name)
        self.assertFalse(objs)

        # Find Them with new name
        objs = collection.find(first_name=new_name)
        self.assertEqual(len(objs), _count_2)

        # Count of all
        objs_count = collection.count()
        self.assertEqual(objs_count, _count_1 + _count_2)

    # Fields
    def test_document_fields(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Find
        obj = collection.find_one(first_name=first_name, last_name=last_name)

        self.assertEqual(set(obj.data.keys()), {'first_name', 'last_name', '_id'})

    # Save
    def test_document_save_method(self):
        collection = self.db.collection(f.word())

        # Insert with specific name
        first_name = f.first_name()
        collection.insert_one(first_name=first_name, last_name=f.last_name())

        # Find One
        obj = collection.find_one(first_name=first_name)
        new_name = f.first_name()

        # Update it
        obj.first_name = new_name
        obj.save()

        self.assertEqual(obj.first_name, new_name)

        # Find with old name
        old_obj = collection.find_one(first_name=first_name)
        self.assertIsNone(old_obj)

        # Find with new name
        obj = collection.find_one(first_name=new_name)
        self.assertEqual(obj.first_name, new_name)

    # Json
    def test_document_json_method(self):
        collection = self.db.collection(f.word())
        first_name = f.first_name()
        last_name = f.last_name()

        # Insert with specific names
        collection.insert_one(first_name=first_name, last_name=last_name)

        # Find
        obj = collection.find_one(first_name=first_name, last_name=last_name)

        _json = {
            'first_name': first_name,
            'last_name': last_name,
            '_id': 1,
        }
        self.assertEqual(obj.json(), json.dumps(_json).decode())


# TODO: Test whole scenario with -> secret_key, return_dict
# TODO: Test where exceptions happen



import re

from setuptools import setup


def pantherdb_version() -> str:
    with open('pantherdb/__init__.py') as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


VERSION = pantherdb_version()
with open('README.md') as file:
    DESCRIPTION = file.read()

setup(
    name='pantherdb',
    version=VERSION,
    python_requires='>=3.8',
    author='Ali RajabNezhad',
    author_email='alirn76@yahoo.com',
    url='https://github.com/alirn76/pantherdb',
    description='is a Simple, FileBase and Document Oriented database',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=[
        'orjson~=3.9.5',
        'cryptography~=41.0',
    ],
)
